package ch.psi.daq.rest.queries;

import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;

import org.apache.commons.lang3.builder.ReflectionToStringBuilder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.util.Assert;

import com.fasterxml.jackson.annotation.JsonCreator;
import com.fasterxml.jackson.annotation.JsonProperty;
import com.fasterxml.jackson.annotation.JsonSubTypes;
import com.fasterxml.jackson.annotation.JsonSubTypes.Type;
import com.fasterxml.jackson.annotation.JsonTypeInfo;

import ch.psi.daq.cassandra.reader.Ordering;
import ch.psi.daq.hazelcast.query.Aggregation;
import ch.psi.daq.hazelcast.query.AggregationType;
import ch.psi.daq.hazelcast.query.Query;
import ch.psi.daq.hazelcast.query.bin.BinningStrategy;
import ch.psi.daq.hazelcast.query.bin.BinningStrategyFactory;
import ch.psi.daq.hazelcast.query.range.QueryRange;

/**
 * 
 * @author zellweger_c
 *
 */
@JsonTypeInfo(
      use = JsonTypeInfo.Id.NAME,
      include = JsonTypeInfo.As.PROPERTY,
      property = "queryType")
@JsonSubTypes(
{
      @Type(value = PulseRangeQuery.class, name = "pulserange"),
      @Type(value = TimeRangeQuery.class, name = "timerange"),
})
public abstract class AbstractQuery implements Query {

   private static Logger logger = LoggerFactory.getLogger(AbstractQuery.class);

   private List<String> channels;

   private LinkedHashSet<String> fields;

   private List<Aggregation> aggregations;

   private AggregationType aggregationType;

   private boolean aggregateChannels;

   private Ordering ordering;

   private BinningStrategy binningStrategy;

   private BinningStrategyEnum binningStrategyEnum;

   private QueryRange queryRange;

   /**
    * Constructor.
    * 
    * @param ordering whether to add a 'orderBy' clause into the database query
    * @param channels all the channelIds (channel names) we want to query
    * @param fields the fields (who map to fields in the DB) we are interested in returning to the
    *        client, needs to be in insertion order (hence the {@link LinkedHashSet} type)
    * @param binningStrategyEnum enum that maps the user's String to a concrete
    *        {@link BinningStrategy} implementation
    * @param binLengthOrCount depending on the chosen binning strategy, this field defines either the
    *        count (how many pulse ids are to be put inside 1 bin) or the time frame for 1 bin
    * @param aggregateChannels whether aggregation will include all channels, default is on a
    *        per-channel basis
    * @param aggregationType defines whether aggregation takes place in an index- or value-based
    *        manner
    * @param aggregations list of aggregations / statistics to calculate, e.g. min, max and average
    * @param queryRange object containing the ranges for either pulse-based queries or time-based
    *        queries
    */
   @JsonCreator
   public AbstractQuery(
         // note that those annotations are needed for the polymorphic
         // mapping to work correctly
         @JsonProperty(value = "ordering") Ordering ordering,
         @JsonProperty(value = "channels") List<String> channels,
         @JsonProperty(value = "fields") LinkedHashSet<String> fields,
         @JsonProperty(value = "binningStrategy") BinningStrategyEnum binningStrategyEnum,
         @JsonProperty(value = "binDuration") long binLengthOrCount,
         @JsonProperty(value = "aggregateChannels") boolean aggregateChannels,
         @JsonProperty(value = "aggregationType") AggregationType aggregationType,
         @JsonProperty(value = "aggregations") List<Aggregation> aggregations,
         @JsonProperty(value = "queryRange") QueryRange queryRange) {

      this.ordering = ordering;
      this.aggregateChannels = aggregateChannels;
      this.aggregationType = aggregationType;
      this.aggregations = aggregations;
      this.queryRange = queryRange;

      Assert.notNull(channels, "channel name must not be null.");
      Assert.notNull(fields, "field, i.e. property, names must not be null.");

      this.channels = channels;
      this.fields = fields;

      this.binningStrategyEnum = binningStrategyEnum; // can be null: default then will be
                                                      // BinCountBinningStrategy


      if (binningStrategyEnum != null) {
         switch (binningStrategyEnum) {
            case count:
               this.binningStrategy = BinningStrategyFactory.getBinningStrategy(getQueryRange(), (int) binLengthOrCount);
               break;
            case length:
               this.binningStrategy = BinningStrategyFactory.getBinningStrategy(getQueryRange(), binLengthOrCount);
               break;
            default:
               logger.warn("No binning strategy has been set. Selecting BinningStrategyBinCount.");
               this.binningStrategy = BinningStrategyFactory.getBinningStrategy(getQueryRange(), (int) binLengthOrCount);
         }
      } else {
         this.binningStrategy = BinningStrategyFactory.getBinningStrategy(getQueryRange(), (int) binLengthOrCount);
      }
   }

   /**
    * {@inheritDoc}
    */
   @Override
   public List<String> getChannels() {
      return Collections.unmodifiableList(channels);
   }

   /**
    * {@inheritDoc}
    */
   @Override
   public Ordering getOrdering() {
      return ordering;
   }

   /**
    * {@inheritDoc}
    */
   @Override
   public AggregationType getAggregationType() {
      return aggregationType;
   }


   @Override
   public boolean isAggregateChannels() {
      return aggregateChannels;
   }

   public BinningStrategyEnum getBinningStrategyEnum() {
      return binningStrategyEnum;
   }

   @Override
   public BinningStrategy getBinningStrategy() {
      return binningStrategy;
   }

   public LinkedHashSet<String> getFields() {
      return fields;
   }

   public List<Aggregation> getAggregations() {
      return aggregations;
   }

   /**
    * {@inheritDoc}
    */
   @Override
   public QueryRange getQueryRange() {
      return queryRange;
   }

   public void setQueryRange(QueryRange queryRange) {
      this.queryRange = queryRange;
   }

   @Override
   public String toString() {
      return ReflectionToStringBuilder.reflectionToString(this);
   }

}
